/*
** $Id: loslib.c,v 1.19 2006/04/26 18:19:49 roberto Exp $
** Standard Operating System library
** See Copyright Notice in agena.h
*/


#include <errno.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h> /* for mkdir, stat */

#if _WIN32
#include <windows.h>  /* for os_memstate */
#include <io.h>       /* for access */
#endif

/* This OS2 header must be placed before including unistd, utsname, ncurses,
   otherwise GCC will not compile successfully. 0.13.3 */
#ifdef __OS2__
#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_NOPM
#define INCL_DOSPROCESS  /* for DosBeep */
#include <os2.h>     /* for memory status queries */
#endif

#if defined(__unix__) || defined(__OS2__)
#include <unistd.h>   /* for access and sleep */
#include <sys/utsname.h>  /* for uname */
#endif


#include <dirent.h>  /* used by io_ls function */
#include <errno.h>   /* used by io_ls function */

#include "chelpers.h"

#define loslib_c
#define LUA_LIB

#include "agena.h"

#include "aauxlib.h"
#include "agenalib.h"


static int os_pushresult (lua_State *L, int i, const char *filename) {
  int en = errno;  /* calls to Lua API may change this value */
  if (i) {
    lua_pushtrue(L);
    return 1;
  }
  else {
    lua_pushfail(L);
    lua_pushfstring(L, "%s: %s", filename, strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
}


static int os_execute (lua_State *L) {
  lua_pushinteger(L, system(luaL_optstring(L, 1, NULL)));
  return 1;
}


static int os_rm (lua_State *L) {
  const char *filename = luaL_checkstring(L, 1);
  return os_pushresult(L, remove(filename) == 0, filename);
}


static int os_rename (lua_State *L) {
  const char *fromname = luaL_checkstring(L, 1);
  const char *toname = luaL_checkstring(L, 2);
  return os_pushresult(L, rename(fromname, toname) == 0, fromname);
}


static int os_tmpname (lua_State *L) {
  char buff[LUA_TMPNAMBUFSIZE];
  int err;
  lua_tmpnam(buff, err);
  if (err)
    return luaL_error(L, "unable to generate a unique filename");
  lua_pushstring(L, buff);
  return 1;
}


static int os_getenv (lua_State *L) {
  lua_pushstring(L, getenv(luaL_checkstring(L, 1)));  /* if NULL push nil */
  return 1;
}


/*
** {======================================================
** Time/Date operations
** { year=%Y, month=%m, day=%d, hour=%H, min=%M, sec=%S,
**   wday=%w+1, yday=%j, isdst=? }
** =======================================================
*/

static void setfield (lua_State *L, const char *key, int value) {
  lua_pushinteger(L, value);
  lua_setfield(L, -2, key);
}

static void setboolfield (lua_State *L, const char *key, int value) {
  if (value < 0)  /* undefined? */
    return;  /* does not set field */
  lua_pushboolean(L, value);
  lua_setfield(L, -2, key);
}

static int getboolfield (lua_State *L, const char *key) {
  int res;
  lua_getfield(L, -1, key);
  res = lua_isnil(L, -1) ? -1 : lua_toboolean(L, -1);
  lua_pop(L, 1);
  return res;
}


static int getfield (lua_State *L, const char *key, int d) {
  int res;
  lua_getfield(L, -1, key);
  if (lua_isnumber(L, -1))
    res = (int)lua_tointeger(L, -1);
  else {
    if (d < 0)
      return luaL_error(L, "field " LUA_QS " missing in date table", key);
    res = d;
  }
  lua_pop(L, 1);
  return res;
}


static int os_date (lua_State *L) {
  const char *s = luaL_optstring(L, 1, "%c");
  time_t t = luaL_opt(L, (time_t)luaL_checknumber, 2, time(NULL));
  struct tm *stm;
  if (*s == '!') {  /* UTC? */
    stm = gmtime(&t);
    s++;  /* skip `!' */
  }
  else
    stm = localtime(&t);
  if (stm == NULL)  /* invalid date? */
    lua_pushnil(L);
  else if (strcmp(s, "*t") == 0) {
    lua_createtable(L, 0, 9);  /* 9 = number of fields */
    setfield(L, "sec", stm->tm_sec);
    setfield(L, "min", stm->tm_min);
    setfield(L, "hour", stm->tm_hour);
    setfield(L, "day", stm->tm_mday);
    setfield(L, "month", stm->tm_mon+1);
    setfield(L, "year", stm->tm_year+1900);
    setfield(L, "wday", stm->tm_wday+1);
    setfield(L, "yday", stm->tm_yday+1);
    setboolfield(L, "isdst", stm->tm_isdst);
  }
  else {  /* Lua 5.1.2 patch */
    char cc[3];
    luaL_Buffer b;
    cc[0] = '%'; cc[2] = '\0';
    luaL_buffinit(L, &b);
    for (; *s; s++) {
      if (*s != '%' || *(s + 1) == '\0')  /* no conversion specifier? */
        luaL_addchar(&b, *s);
      else {
        size_t reslen;
        char buff[200];  /* should be big enough for any conversion result */
        cc[1] = *(++s);
        reslen = strftime(buff, sizeof(buff), cc, stm);
        luaL_addlstring(&b, buff, reslen);
      }
    }
    luaL_pushresult(&b);
  }
  return 1;
}


static int os_time (lua_State *L) {
  time_t t;
  if (lua_isnoneornil(L, 1))  /* called without args? */
    t = time(NULL);  /* get current time */
  else {
    struct tm ts;
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_settop(L, 1);  /* make sure table is at the top */
    ts.tm_sec = getfield(L, "sec", 0);
    ts.tm_min = getfield(L, "min", 0);
    ts.tm_hour = getfield(L, "hour", 12);
    ts.tm_mday = getfield(L, "day", -1);
    ts.tm_mon = getfield(L, "month", -1) - 1;
    ts.tm_year = getfield(L, "year", -1) - 1900;
    ts.tm_isdst = getboolfield(L, "isdst");
    t = mktime(&ts);
  }
  if (t == (time_t)(-1))
    lua_pushnil(L);
  else
    lua_pushnumber(L, (lua_Number)t);
  return 1;
}


static int os_difftime (lua_State *L) {
  lua_pushnumber(L, difftime((time_t)(luaL_checknumber(L, 1)),
                             (time_t)(luaL_optnumber(L, 2, 0))));
  return 1;
}

/* }====================================================== */


static int os_setlocale (lua_State *L) {
  static const int cat[] = {LC_ALL, LC_COLLATE, LC_CTYPE, LC_MONETARY,
                      LC_NUMERIC, LC_TIME};
  static const char *const catnames[] = {"all", "collate", "ctype", "monetary",
     "numeric", "time", NULL};
  const char *l = luaL_optstring(L, 1, NULL);
  int op = luaL_checkoption(L, 2, "all", catnames);
  lua_pushstring(L, setlocale(cat[op], l));
  return 1;
}


static int os_exit (lua_State *L) {
  exit(luaL_optint(L, 1, EXIT_SUCCESS));
  return 0;  /* to avoid warnings */
}


/* determine information on available memory on Windows platforms, 0.5.3, September 15, 2007;
   extended in 0.12.2 on September 29, 2008 to work in UNIX; extended to fully support OS/2 in 0.13.3 */

static int os_memstate (lua_State *L) {
#if defined(_WIN32) || (defined(__unix__) && !defined(LUA_DOS)) || defined(__OS2__)
  long AvailPhys, TotalPhys;
  lua_Number div;
#if defined(_WIN32)
  MEMORYSTATUS memstat;
  long AvailVirtual, TotalVirtual, MemoryLoad;
  int Length;
#elif __OS2__
  int result;
  ULONG memInfo[3];
#endif
  const char *option = luaL_optstring(L, 1, "bytes");
  if (strcmp(option, "bytes") == 0)
    div = 1;
  else if (strcmp(option, "kbytes") == 0)
    div = 1024;
  else if (strcmp(option, "mbytes") == 0)
    div = 1024*1024;
  else if (strcmp(option, "gbytes") == 0)
    div = 1024*1024*1024;
  else
    return luaL_error(L, "unknown option %s", option);
#if defined(_WIN32)
  memstat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memstat);
  AvailPhys = (long)memstat.dwAvailPhys;
  TotalPhys = (long)memstat.dwTotalPhys;
  AvailVirtual = (long)memstat.dwAvailVirtual;
  TotalVirtual = (long)memstat.dwTotalVirtual;
  MemoryLoad = (long)memstat.dwMemoryLoad;
  Length = (int)memstat.dwLength;
  lua_newtable(L);
  lua_pushstring(L, "freephysical");
  lua_pushnumber(L, (lua_Number)(AvailPhys)/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "totalphysical");
  lua_pushnumber(L, (lua_Number)(TotalPhys)/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "freevirtual");
  lua_pushnumber(L, (lua_Number)(AvailVirtual)/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "totalvirtual");
  lua_pushnumber(L, (lua_Number)(TotalPhys)/div);
  lua_rawset(L, -3);
#elif defined(__unix__) && !defined(LUA_DOS)  /* for DJGPP */
  AvailPhys = sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE);
  TotalPhys = sysconf(_SC_PHYS_PAGES) * sysconf(_SC_PAGESIZE);
  lua_newtable(L);
  lua_pushstring(L, "freephysical");
  lua_pushnumber(L, (lua_Number)AvailPhys/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "totalphysical");
  lua_pushnumber(L, (lua_Number)TotalPhys/div);
  lua_rawset(L, -3);
#elif __OS2__
  result = DosQuerySysInfo(QSV_TOTPHYSMEM, QSV_TOTAVAILMEM,
      &memInfo[0], sizeof(memInfo));
  AvailPhys = 0;
  TotalPhys = 0;
    /* sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE); does not work for
      the _SC* names are not defined in GCC unistd.h */
  if (result != 0) {
    lua_pushfail(L);
    return 1;
  }
  lua_newtable(L);
  lua_pushstring(L, "totalphysical");
  lua_pushnumber(L, (lua_Number)memInfo[0]/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "freevirtual");
  lua_pushnumber(L, (lua_Number)memInfo[2]/div);
  lua_rawset(L, -3);
  lua_pushstring(L, "resident");
  lua_pushnumber(L, (lua_Number)memInfo[1]/div);
  lua_rawset(L, -3);
#endif
#else
  lua_pushfail(L);
#endif
  return 1;
}

/* get the current Windows login name, 0.5.3, September 15, 2007;
   extended 0.12.2, October 12, 2008; changed to support OS/2 0.13.3 */

static int os_login (lua_State *L) {
#if defined(_WIN32)
  char *LoginName;
  char buffer[200];
  DWORD len = 200;
  if (GetUserName(buffer, &len)) {
    LoginName = (char *)malloc(len+1);
    if (!LoginName)
      return luaL_error(L, "could not determine login name.\n");
    strncpy(LoginName, buffer, len);
    LoginName[len] = '\0';
  }
  else {
    LoginName = "<Unknown>";
  }
  lua_pushstring(L, LoginName);
#elif defined(__unix__) || defined(__OS2__)
  char *LoginName;
  LoginName = getlogin();
  if (LoginName == NULL)
    lua_pushfail(L);
  else
    lua_pushstring(L, LoginName);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* get the Windows computer name, 0.5.3, September 15, 2007;
   improved in 0.13.3 */

static int os_computername (lua_State *L) {
#if defined(_WIN32)
  char *NTComputerName;
  char buffer[256];
  DWORD len = 256;
  if (GetComputerName(buffer, &len)) {
    NTComputerName = (char *)malloc(len+1);
    if (!NTComputerName)
      return luaL_error(L, "could not determine login name.\n");
    strncpy(NTComputerName, buffer, len);
    NTComputerName[len] = '\0';
  }
  else {
    NTComputerName = "<Unknown>";
  }
  lua_pushstring(L, NTComputerName);
#elif defined(__unix__)
  /* does not work with OS/2: needs a file `TCPIP32` (as does bash). If not
     present, Agena does not start. */
  char buffer[256];
  size_t len = 256;
  if (gethostname(buffer, len) == 0)
    lua_pushstring(L, buffer);
  else
    lua_pushfail(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* return amount of free memory; extended 0.12.2, 28.09.2008 */
static int os_freemem (lua_State *L) {
#if defined(_WIN32) || (defined(__unix__) && !defined(LUA_DOS))
  lua_Number div;
  const char *option = luaL_optstring(L, 1, "bytes");
#if defined(_WIN32)
  MEMORYSTATUS memstat;
#endif
  if (strcmp(option, "bytes") == 0)
    div = 1;
  else if (strcmp(option, "kbytes") == 0)
    div = 1024;
  else if (strcmp(option, "mbytes") == 0)
    div = 1024*1024;
  else if (strcmp(option, "gbytes") == 0)
    div = 1024*1024*1024;
  else
    return luaL_error(L, "unknown option %s", option);
#if defined(_WIN32)
  memstat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memstat);
  lua_pushnumber(L, (lua_Number)memstat.dwAvailPhys/div);
#elif defined(__unix__) && !defined(LUA_DOS) /* excludes DJGPP */
  lua_pushnumber(L, (lua_Number)(sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE))/div);
#endif
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_lscore (lua_State *L) {
  DIR *dir;
  struct dirent *entry;
  int i, simple;
  const char *path;
  path = luaL_optstring(L, 1, tools_cwd());
  simple = (lua_gettop(L) < 2);
  dir = opendir(path);
  if (dir == NULL) {
    lua_pushnil(L);
    lua_pushstring(L, strerror(errno));
    return 2;
  }
  lua_newtable(L);
  i = 1;
  while ((entry = readdir(dir)) != NULL) {
    if (simple && ( (strcmp(entry->d_name, ".") == 0) || (strcmp(entry->d_name, "..") == 0) )) continue;
    lua_pushnumber(L, i++);
    lua_pushstring(L, entry->d_name);
    lua_settable(L, -3);
  }
  closedir(dir);
  return 1;
}


static int os_fstat (lua_State *L) {  /* 0.12.2, October 11, 2008 */
  struct stat entry;
  int i;
  const char *fn;
  mode_t mode;
  struct tm *stm;
  time_t seconds;
  fn = luaL_checkstring(L, 1);
  lua_newtable(L);
  i = 1;
  if (stat(fn, &entry) == 0) {
    mode = entry.st_mode;
    lua_pushnumber(L, i++);
    if (S_ISDIR(mode))
      lua_pushstring(L, "DIR");
#if defined(__unix__) && !defined(LUA_DOS)
    else if (S_ISLNK(mode))
      lua_pushstring(L, "LINK");
#endif
    else if (S_ISREG(mode))
      lua_pushstring(L, "FILE");
    else if (S_ISCHR(mode))
      lua_pushstring(L, "CHARSPECFILE");
    else if (S_ISBLK(mode))
      lua_pushstring(L, "BLOCKSPECFILE");
    else
      lua_pushstring(L, "OTHER");
    lua_settable(L, -3);
    lua_pushnumber(L, i++);
    lua_pushnumber(L, entry.st_size);
    lua_settable(L, -3);
    seconds = entry.st_mtime;
    stm = localtime(&seconds);
    lua_pushnumber(L, i++);
    lua_createtable(L, 6, 0);  /* 6 = number of entries */
    lua_rawsetinumber(L, -1, 1, stm->tm_year+1900);
    lua_rawsetinumber(L, -1, 2, stm->tm_mon+1);
    lua_rawsetinumber(L, -1, 3, stm->tm_mday);
    lua_rawsetinumber(L, -1, 4, stm->tm_hour);
    lua_rawsetinumber(L, -1, 5, stm->tm_min);
    lua_rawsetinumber(L, -1, 6, stm->tm_sec);
    lua_settable(L, -3);
  }
  return 1;
}


static int os_fexists (lua_State *L) {
  lua_pushboolean(L, (access(luaL_checkstring(L, 1), 00) != -1));
  return 1;
}


/* print working directory; July 07, 2008 */
static int os_pwd (lua_State *L) {
  char *buffer = malloc(sizeof(char)*256);
  getcwd(buffer, 256);
  if (buffer == NULL)
    lua_pushfail(L);
  else
    lua_pushstring(L, buffer);
  free(buffer);
  return 1;
}


/* The following routines were taken from the lposix.c file in the
*  POSIX library for Lua 5.0. Based on original by Claudio Terra for Lua 3.x.
*  Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
*  05 Nov 2003 22:09:10
*/

/** mkdir(path), modified for Agena; extended in 0.13.3 to support OS/2 */
static int os_md (lua_State *L) {
#ifdef _WIN32
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, mkdir(path), path);
#elif defined(__unix__) || defined(__OS2__)
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, mkdir(path, 0777), path);
#else
  lua_pushfail(L);
  return 1;
#endif
}

/* chdir(path) */
static int os_cd (lua_State *L) {
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, chdir(path), path);
}

/* rmdir(path) */
static int os_rd (lua_State *L) {
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, rmdir(path), path);
}


/* beep: sound the loudspeaker with a given frequency (first arg, a posint) and
   a duration (second arg, in seconds, a float), April 01, 2007 */

static int os_beep (lua_State *L) {
#ifdef _WIN32
  int nargs = lua_gettop(L);
  if (nargs == 0) {
    putc('\007', stderr);
    lua_pushnil(L);
  } else {
    int tone;
    lua_Number duration;
    tone = agn_checknumber(L, 1);
    duration = agn_checknumber(L, 2);
    if (tone > 0 && duration > 0) {
      Beep(tone, duration*1000);
      lua_pushnil(L);
    } else
      lua_pushfail(L);
  }
#elif defined(__OS2__)
  int nargs = lua_gettop(L);
  if (nargs == 0) {
    putc('\007', stderr);
    lua_pushnil(L);
  } else {
    ULONG tone;
    lua_Number duration;
    tone = agn_checknumber(L, 1);
    duration = agn_checknumber(L, 2);
    if (tone > 0 && duration > 0) {
      if (DosBeep(tone, (ULONG)(duration*1000)))  /* invalid frequency */
        lua_pushfail(L);
      else
        lua_pushnil(L);
    } else
      lua_pushfail(L);
  }
#elif defined(__unix__)
  putc('\007', stderr);  /* FIXME: does not work on Sparc */
  lua_pushnil(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* wait: wait for x seconds; March 31, 2007 */

static int os_wait (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  if (x > 0) {
    #ifdef _WIN32
    Sleep(x*1000);
    lua_pushnil(L);
    #elif defined(__unix__) || defined(__OS2__)
    sleep(x);
    lua_pushnil(L);
    #else
    lua_pushfail(L);
    #endif
  } else lua_pushfail(L);
  return 1;
}


/* retrieve information on your platform;
0.12.2, October 12, 2008; extended to correctly identify Vista, Windows 2003, and Windows 2007 Server
on April 08, 2009 - 0.13.4 */

#ifdef _WIN32

typedef struct WinVer {
  int BuildNumber;
  int MajorVersion;
  int MinorVersion;
  int PlatformId;
  char ProductType;
} WinVer;


int getWindowsVersion (struct WinVer *winver) {
   int result, MajorVersion, MinorVersion, ProductType, PlatformId;
   OSVERSIONINFOEX version;
   version.dwOSVersionInfoSize = sizeof(version);
   GetVersionEx((OSVERSIONINFO *)&version);

   /* dwMajorVersion
   Major version number of the operating system. This member can be one of the following values. Value Meaning
   4 Windows NT 4.0, Windows Me, Windows 98, or Windows 95.
   5 Windows Server 2003, Windows XP, or Windows 2000
   6 Windows Vista, Windows Server 2008, Windows 7

   dwMinorVersion
   Minor version number of the operating system. This member can be one of the following values. Value Meaning
   0 Windows 2000, Windows NT 4.0, Windows 95, or Windows Vista
   1 Windows XP, Windows 2008 server
   2 Windows Server 2003, Windows 7
   10 Windows 98
   90 Windows Me

   dwPlatformId
   Operating system platform. This member can be one of the following values. Value Meaning
   VER_PLATFORM_WIN32_NT
   2 Windows Server 2003, Windows XP, Windows 2000, or Windows NT.
   VER_PLATFORM_WIN32_WINDOWS
   1 Windows Me, Windows 98, or Windows 95.
   VER_PLATFORM_WIN32s
   0 Win32s on Windows 3.1. */

   winver->BuildNumber = (int)version.dwBuildNumber;
   MajorVersion = winver->MajorVersion = (int)version.dwMajorVersion;
   MinorVersion = winver->MinorVersion = (int)version.dwMinorVersion;
   PlatformId = winver->PlatformId = (int)version.dwPlatformId;
   ProductType = winver->ProductType = (char)version.wProductType;
   if (MajorVersion == 4 && MinorVersion == 0 && PlatformId == 1)
      result = 2;  /* Win 95 */
   else if (MajorVersion == 4 && MinorVersion == 0 && PlatformId == 2)
      result = 3;  /* Win NT 4.0 */
   else if (MajorVersion == 4 && MinorVersion == 10)
      result = 4;  /* Win 98 */
   else if (MajorVersion == 5 && MinorVersion == 0)
      result = 5;  /* Win 2000 */
   else if (MajorVersion == 4 && MinorVersion == 90)
      result = 6;  /* Win Me */
   else if (MajorVersion == 5 && MinorVersion == 1)
      result = 7;  /* Win XP */
   else if (MajorVersion == 6 && MinorVersion == 0 && ProductType == VER_NT_WORKSTATION)
      result = 8;  /* Win Vista */
   else if (MajorVersion == 6 && MinorVersion >= 0 && ProductType != VER_NT_WORKSTATION)
      result = 9;  /* Win Server 2008 */
   else if (MajorVersion == 5 && MinorVersion == 2)
      result = 10; /* Win 2003 */
   else if (MajorVersion == 6 && MinorVersion == 1 && ProductType == VER_NT_WORKSTATION)
      result = 11; /* Windows 7 */
   else if (PlatformId == 0)
      result = 1;  /* Win32s under Windows 3.1 */
   else
      result = 0;
   return result;
}
#endif


/* Novell DOS 7 = 'IBMPcDos', '6', '00', 'pc'
   OS/2 Warp 4 Fixpack 15 = 'OS/2', '1', '2.45', 'i386'
   Windows = 'Windows', etc. */

static int os_system (lua_State *L) {
#ifdef _WIN32
   /* dwMajorVersion
   Major version number of the operating system. This member can be one of the following values. Value Meaning
   4 Windows NT 4.0, Windows Me, Windows 98, or Windows 95.
   5 Windows Server 2003, Windows XP, or Windows 2000
   6 Windows Vista, Windows Server 2008, Windows 7

   dwMinorVersion
   Minor version number of the operating system. This member can be one of the following values. Value Meaning
   0 Windows 2000, Windows NT 4.0, Windows 95, or Windows Vista
   1 Windows XP, Windows 2008 server
   2 Windows Server 2003, Windows 7
   10 Windows 98
   90 Windows Me

   dwPlatformId
   Operating system platform. This member can be one of the following values. Value Meaning
   VER_PLATFORM_WIN32_NT
   2 Windows Server 2003, Windows XP, Windows 2000, or Windows NT.
   VER_PLATFORM_WIN32_WINDOWS
   1 Windows Me, Windows 98, or Windows 95.
   VER_PLATFORM_WIN32s
   0 Win32s on Windows 3.1. */

  int winver;
  struct WinVer winversion;
  static char *versions[] = {  /* extended 0.13.4 */
    "unknown", "32s", "95", "NT 4.0", "98", "2000", "ME", "XP", "Vista", "Server 2008", "Server 2003", "7"
  };
  winver = getWindowsVersion(&winversion);
  lua_newtable(L);
  lua_rawsetistring(L, -1, 1, "Windows");
  lua_rawsetistring(L, -1, 2, versions[winver]);
  lua_rawsetinumber(L, -1, 3, winversion.BuildNumber);
  lua_rawsetinumber(L, -1, 4, winversion.PlatformId);
  lua_rawsetinumber(L, -1, 5, winversion.MajorVersion);  /* added 0.13.4 */
  lua_rawsetinumber(L, -1, 6, winversion.MinorVersion);  /* added 0.13.4 */
  lua_rawsetinumber(L, -1, 7, winversion.ProductType);   /* added 0.13.4 */
#elif defined(__unix__) || defined(__OS2__)
  /* applies to DJGPP compiled DOS code as well */
  struct utsname info;
  if (uname(&info) != -1) {
    lua_newtable(L);
    lua_rawsetistring(L, -1, 1, info.sysname);
    lua_rawsetistring(L, -1, 2, info.release);
    lua_rawsetistring(L, -1, 3, info.version);
    lua_rawsetistring(L, -1, 4, info.machine);
   }
   else
     lua_pushfail(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_endian (lua_State *L) {
  int r = tools_endian();
  if (r == -1)
    lua_pushfail(L);
  else
    lua_pushinteger(L, r);
  return 1;
}


/* return status of battery on laptops, works in Windows 2000 and higher. 0.14.0, March 13, 2009 */
static int os_battery (lua_State *L) {
#ifdef _WIN32
  BYTE bACLineStatus, status, bBatteryLifePercent, bReserved1;
  DWORD bBatteryLifeTime, bBatteryFullLifeTime;
  SYSTEM_POWER_STATUS battstat;
  int BatteryCharging, winver;
  struct WinVer windowsver;
  static char *ACstatus[] = {"off", "on", "unknown"};
  static char *BatteryStatusMapping[] = {"medium", "high", "low", "critical", "charging", "no battery", "unknown"};
  winver = getWindowsVersion(&windowsver);
  if (winver < 2 || winver == 3) {  /* Windows NT 4.0 and earlier are not supported */
    lua_pushfail(L);
    return 1;
  }
  /* get battery information by calling the Win32 API */
  GetSystemPowerStatus(&battstat);
  bReserved1 = battstat.Reserved1;
  bACLineStatus = battstat.ACLineStatus;
  BatteryCharging = 0; /* not charging */
  if (bACLineStatus == 255)
    bACLineStatus = 2;
  switch (battstat.BatteryFlag) {
    case 0: /* >= 33 and <= 66 % */
      status = 0;
      break;
    case 1: /* high */
      status = 1;
      break;
    case 2: /* low */
      status = 2;
      break;
    case 4: /* critical */
      status = 3;
      break;
    case 8: case 9: case 10:  /* charging */
      status = battstat.BatteryFlag - 8;
      BatteryCharging = 1;
      break;
    case 12:  /* charging & critical */
      status = 3;
      BatteryCharging = 1;
    case 128: /* No system battery */
      status = 5;
      break;
    case 255: /* unknown status */
      status = 6;
      break;
    default:  /* got value 0 at around 45% on Acer, 10 right after reconnecting AC line  */
      status = 6;
      break;
  }
  bBatteryLifePercent = battstat.BatteryLifePercent;
  bBatteryLifeTime = battstat.BatteryLifeTime;
  bBatteryFullLifeTime = battstat.BatteryFullLifeTime;
  lua_createtable(L, 0, 0);
  lua_pushstring(L, "acline");
  lua_pushstring(L, ACstatus[bACLineStatus]);
  lua_settable(L, -3);
  if (status == 4) {
    lua_pushstring(L, "installed");
    lua_pushfalse(L);
    lua_settable(L, -3);
  } else {
    lua_pushstring(L, "installed");
    lua_pushtrue(L);
    lua_settable(L, -3);
    lua_pushstring(L, "life");
    lua_pushnumber(L, bBatteryLifePercent);
    lua_settable(L, -3);
    lua_pushstring(L, "status");
    lua_pushstring(L, BatteryStatusMapping[status]);
    lua_settable(L, -3);
    lua_pushstring(L, "charging");
    lua_pushboolean(L, (BatteryCharging==0));
    lua_settable(L, -3);
    lua_pushstring(L, "flag");
    lua_pushnumber(L, battstat.BatteryFlag);
    lua_settable(L, -3);
    lua_pushstring(L, "lifetime");
    if (bBatteryLifeTime != -1)
      lua_pushnumber(L, bBatteryLifeTime);
    else
      lua_pushundefined(L);
    lua_settable(L, -3);
    lua_pushstring(L, "fulllifetime");
    if (bBatteryFullLifeTime != -1)
      lua_pushnumber(L, bBatteryFullLifeTime);
    else
      lua_pushundefined(L);
    lua_settable(L, -3);
  }
#elif __OS2__
#define IOCTL_POWER 0x000c
#define POWER_GETPOWERSTATUS 0x0060
  int result;
  HFILE pAPI;
  ULONG call;
  struct _Battery {
    USHORT usParamLen;
    USHORT usPowerFlags;
    UCHAR  ucACStatus;
    UCHAR  ucBatteryStatus;
    UCHAR  ucBatteryLife;
  } Battery;
  struct _Data {
    USHORT usReturn;
  } Data;
  ULONG sizeBattery, sizeData;
  result = DosOpen("\\DEV\\APM$", &pAPI, &call, 0,
    FILE_NORMAL, FILE_OPEN, OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE,
    NULL);
  if (result != 0) {
    lua_pushfail(L);
    return 1;
  }
  sizeBattery = Battery.usParamLen = sizeof(Battery);
  sizeData = sizeof(Data);
  result = DosDevIOCtl(pAPI,
    IOCTL_POWER, POWER_GETPOWERSTATUS, (PVOID)&Battery, sizeof(Battery),
    &sizeBattery, (PVOID)&Data, sizeof(Data), &sizeData);
  if (result == ERROR_INVALID_PARAMETER) {
    lua_pushfail(L);
  } else {
    int acstatus, batterystatus;
    static char *ACstatus[] = {"off", "on", "unknown", "invalid"};
    static char *BatteryStatusMapping[] =
      {"high", "low", "critical", "charging", "unknown", "invalid"};
    lua_newtable(L);
    switch (Battery.ucACStatus) {
      case   0: acstatus = 0; break;
      case   1: acstatus = 1; break;
      case 255: acstatus = 2; break;
      default:  acstatus = 3; break;
    }
    switch (Battery.ucBatteryStatus) {
      case 0:   batterystatus = 0; break;
      case 1:   batterystatus = 1; break;
      case 2:   batterystatus = 2; break;
      case 3:   batterystatus = 3; break;
      case 255: batterystatus = 4; break;
      default:  batterystatus = 5; break;
    }
    lua_pushstring(L, "acline");
    lua_pushstring(L, ACstatus[acstatus]);
    lua_rawset(L, -3);
    lua_pushstring(L, "status");
    lua_pushstring(L, BatteryStatusMapping[batterystatus]);
    lua_rawset(L, -3);
    lua_pushstring(L, "flags");
    lua_pushnumber(L, (lua_Number)Battery.usPowerFlags);
    lua_rawset(L, -3);
    lua_pushstring(L, "powermanagement");
    lua_pushboolean(L, ((Battery.usPowerFlags & 0x0001) == 1));
    lua_rawset(L, -3);
    lua_pushstring(L, "life");
    if (Battery.ucACStatus != 1)  /* AC line is off */
      lua_pushnumber(L, (lua_Number)Battery.ucBatteryLife);
    else
      lua_pushundefined(L);
    lua_rawset(L, -3);
  }
#else
  lua_pushfail(L);
#endif
  return 1;
}


static const luaL_Reg syslib[] = {
  {"date",      os_date},
  {"difftime",  os_difftime},
  {"execute",   os_execute},          /* changed in 0.5.4 */
  {"exit",      os_exit},
  {"getenv",    os_getenv},
  {"rm",        os_rm},
  {"rename",    os_rename},
  {"setlocale", os_setlocale},
  {"time",      os_time},
  {"tmpname",   os_tmpname},
  {"memstate",  os_memstate},         /* 0.5.3  Sept 15, 2007 */
  {"computername", os_computername},  /* 0.5.3  Sept 15, 2007 */
  {"system",    os_system},           /* 0.12.2  October 12, 2008 */
  {"login",     os_login},            /* 0.5.3  Sept 15, 2007 */
  {"freemem",   os_freemem},          /* 0.5.3  Sept 15, 2007 */
  {"lscore",    os_lscore},           /* added on August 01, 2007 */
  {"fexists",   os_fexists},          /* added on April 09, 2007 */
  {"fstat",     os_fstat},            /* added on October 11, 2008 */
  {"md",        os_md},               /* added July 07, 2008 */
  {"cd",        os_cd},               /* added July 07, 2008 */
  {"rd",        os_rd},               /* added July 07, 2008 */
  {"pwd",       os_pwd},              /* added July 07, 2008 */
  {"beep",      os_beep},             /* added on March 31, 2007 */
  {"wait",      os_wait},             /* added on March 31, 2007 */
  {"endian",    os_endian},           /* 0.8.0, added on December 11, 2007 */
  {"battery",   os_battery},          /* 0.14.0, added on March 13, 2009 */
  {NULL, NULL}
};

/* }====================================================== */



LUALIB_API int luaopen_os (lua_State *L) {
  luaL_register(L, LUA_OSLIBNAME, syslib);
  return 1;
}


