/*
** $Id: lstatlib.c,v 1.0.0 by Alexander Walz - initiated June 19, 2007
** Statistics library
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>
#include <string.h>

/* the following package ini declarations must be included after `#include <` and before `include #` ! */

#define stats_c
#define LUA_LIB

#include "../src/agena.h"
#include "../src/aauxlib.h"
#include "../src/agenalib.h"


#ifndef LUA_DOS
#define AGENA_STATSLIBNAME "stats"
LUALIB_API int (luaopen_stats) (lua_State *L);
#endif


/* median of all values in an _already sorted set_, June 18, 2007; tweaked Jan 13, 2008;
   faster than an Agena implementation */

static int tbl_median (lua_State *L) {
  int n, type;
  type = lua_type(L, 1);
  luaL_argcheck(L, type == LUA_TTABLE || type == LUA_TSEQ, 1, "table or sequence expected");
  n = agn_nops(L, 1);
  div_t middle;
  middle = div(n, 2);
  switch (type) {
    case LUA_TTABLE: {
      if (middle.rem != 0)
        lua_rawgeti(L, 1, middle.quot+1);
      else
        lua_pushnumber(L,
          (agn_getinumber(L, 1, middle.quot) + agn_getinumber(L, 1, middle.quot+1)) / 2);
    }
    case LUA_TSEQ: {
      if (middle.rem != 0)
        lua_seqgeti(L, 1, middle.quot+1);
      else
        lua_pushnumber(L,
          (lua_seqgetinumber(L, 1, middle.quot) + lua_seqgetinumber(L, 1, middle.quot+1)) / 2);
    }
    default: lua_assert(0);
  }
  return 1;
}


/* tbl_minmax: returns the smallest and largest value in a list. If the option
   'sorted' is given, the list is not traversed, instead the first and the last entry
   in the list is returned.
   If the list is empty, NULL is returned. lua_getn must be called at first,
   otherwise lua_rawgeti would return 0 with empty lists. June 20, 2007 */

static int tbl_minmax (lua_State *L) {
  int i, n, type;
  lua_Number min, max, number;
  type = lua_type(L, 1);
  luaL_argcheck(L, type == LUA_TTABLE || type == LUA_TSEQ, 1, "table or sequence expected");
  const char *option = luaL_optstring(L, 2, "default");
  n = agn_nops(L, 1);
  if (n < 2) {
    if (type == LUA_TTABLE)
      lua_newtable(L);
    else
      agn_createseq(L, 0);
    return 1;
  }
  switch (type) {
    case LUA_TTABLE: {
      lua_newtable(L);
      if (strcmp(option, "sorted") == 0) {
        if (n == 0) {
          return 1;
        }
        lua_pushinteger(L, 1);
        lua_rawgeti(L, 1, 1);
        lua_rawset(L, -3);
        lua_pushinteger(L, 2);
        lua_rawgeti(L, 1, n);
        lua_rawset(L, -3);
        return 1;
      }
      min = agn_getinumber(L, 1, 1);
      max = min;
      for (i=2; i<=n; i++) {
        number = agn_getinumber(L, 1, i);
        if (number < min)
          min = number;
        if (number > max)
          max = number;
      }
      /* FIXME: optimize this */
      lua_pushinteger(L, 1);
      lua_pushnumber(L, min);
      lua_rawset(L, -3);
      lua_pushinteger(L, 2);
      lua_pushnumber(L, max);
      lua_rawset(L, -3);
      break;
    }
    case LUA_TSEQ: {
      agn_createseq(L, 2);
      if (strcmp(option, "sorted") == 0) {
        if (n == 0) {
          return 1;
        }
        lua_seqgeti(L, 1, 1);
        lua_seqseti(L, -2, 1);
        lua_seqgeti(L, 1, n);
        lua_seqseti(L, -2, 2);
        return 1;
      }
      min = lua_seqgetinumber(L, 1, 1);
      max = min;
      for (i=2; i<=n; i++) {
        number = lua_seqgetinumber(L, 1, i);
        if (number < min)
          min = number;
        if (number > max)
          max = number;
      }
      /* FIXME: optimize this */
      lua_pushnumber(L, min);
      lua_seqinsert(L, -2);
      lua_pushnumber(L, max);
      lua_seqinsert(L, -2);
      break;
    }
    default: lua_assert(0);
  }
  return 1;
}


static const luaL_Reg statslib[] = {
  {"median", tbl_median},                 /* added on June 19, 2007 */
  {"minmax", tbl_minmax},                 /* added on June 20, 2007 */
  {NULL, NULL}
};


static void createmetatable (lua_State *L) {
  lua_createtable(L, 0, 1);  /* create metatable for strings */
  lua_pushliteral(L, "");  /* dummy string */
  lua_pushvalue(L, -2);
  lua_setmetatable(L, -2);  /* set string metatable */
  lua_pop(L, 1);  /* pop dummy string */
  lua_pushvalue(L, -2);  /* string library... */
  lua_setfield(L, -2, "__index");  /* ...is the __index metamethod */
  lua_pop(L, 1);  /* pop metatable */
}


/*
** Open stats library
*/
LUALIB_API int luaopen_stats (lua_State *L) {
  luaL_register(L, AGENA_STATSLIBNAME, statslib);
  createmetatable(L);
  return 1;
}

/* ====================================================================== */
